/* Copyright (C) 2006 Anton Fedorov <datacompboy@call2ru.com>

   nss_vserver: provide auth on host according to passwd & shadow files
		in vservers.
   You can find info about vserver at http://linux-vserver.org/
   See README for details
   
   Feel free to use code as you want
 */

#include <nss.h>
#include <pthread.h>
#include <sys/types.h>
#include <grp.h>
#include <string.h>
#include <errno.h>
#include "config.h"

#define NSS_DB gr
#define NSS_RES group

#include "commdef.h"

typedef struct ENT_CONTEXT {
  DBOPS_T lastop;
  int calls;
} ENT_CONTEXT_T;

static void context_init(ENT_CONTEXT_T * context)
{
  context->lastop = OP_NOOP;
  context->calls  = 0;
}

static void context_free(ENT_CONTEXT_T * context)
{
  context_init(context);
}

#include "commctx.h"

NSS_STATUS CONCAT3(_nss_vserver_set,NSS_DB,ent) (void)
{
  ENT_CONTEXT_T * context = get_context();
  if (context->lastop!=OP_NOOP) {
	context_free(context);
  }
  context->lastop=OP_SETENT;
  context->calls=0;
  return NSS_STATUS_SUCCESS;
}

NSS_STATUS CONCAT3(_nss_vserver_end,NSS_DB,ent) (void)
{
  ENT_CONTEXT_T * context = get_context();
  context_free(context);
  return NSS_STATUS_SUCCESS;
}

#define min(a,b) (a<b?a:b)
#define FILLBUF \
	  memcpy(buffer, VSGRP "\0x\0\0\0\0\0\0\0\0", min(buflen,sizeof(VSGRP "\0x\0\0\0\0\0\0\0\0"))); /* zero at end copied */ \
	  result->gr_name = buffer; \
	  result->gr_passwd = &buffer[strlenVSGRP+1]; \
	  result->gr_gid = VSGID; \
	  result->gr_mem = (char**)&buffer[strlenVSGRP+4]; \
	    /* We will NOT provide full list of vserver users */

NSS_STATUS CONCAT3(_nss_vserver_get,NSS_DB,ent_r) (struct NSS_RES *result,
								 char *buffer, size_t buflen, int *errnop)
{
  ENT_CONTEXT_T * context = get_context();
  if (!(context->lastop == OP_SETENT || context->lastop==OP_GETENT)) {
	*errnop=EPROTO;
	return NSS_STATUS_UNAVAIL;
  }
  context->lastop = OP_GETENT;
  if (context->calls == 0) {
	if (buflen < strlenVSGRP+3+sizeof(char*)) {
	  *errnop=ERANGE;
	  return NSS_STATUS_TRYAGAIN;
	} else {
	  FILLBUF
	  context->calls++;
	  *errnop=0;
	  return NSS_STATUS_SUCCESS;
	}
  } else {
	  *errnop=ENOENT;
	  return NSS_STATUS_NOTFOUND;
  }
}

NSS_STATUS CONCAT3(_nss_vserver_get,NSS_DB,nam_r) (const char *name,
								 struct NSS_RES * result,
								 char *buffer, size_t buflen, int *errnop)
{
  if (strlen(name)!=strlenVSGRP || memcmp(name, VSGRP, strlenVSGRP)) {
	  *errnop=ENOENT;
	  return NSS_STATUS_NOTFOUND;
  } else {
	if (buflen < strlenVSGRP+3+sizeof(char*)) {
	  *errnop=ERANGE;
	  return NSS_STATUS_TRYAGAIN;
	} else {
	  FILLBUF
	  *errnop=0;
	  return NSS_STATUS_SUCCESS;
	}
  }
}

NSS_STATUS CONCAT3(_nss_vserver_get,NSS_DB,gid_r) (gid_t gid,
									struct NSS_RES *result,
									char *buffer, size_t buflen, int *errnop)
{
  if (gid!=VSGID) {
	*errnop=ENOENT;
	return NSS_STATUS_NOTFOUND;
  } else {
	if (buflen < strlenVSGRP+3+sizeof(char*)) {
	  *errnop=ERANGE;
	  return NSS_STATUS_TRYAGAIN;
	} else {
	  FILLBUF
	  *errnop=0;
	  return NSS_STATUS_SUCCESS;
	}
  }
}
