/* Copyright (C) 2006 Anton Fedorov <datacompboy@call2ru.com>

   nss_vserver: provide auth on host according to passwd & shadow files
		in vservers.
   You can find info about vserver at http://linux-vserver.org/
   See README for details
   
   Feel free to use code as you want
 */


#include <shadow.h>
#include <string.h>
#include "config.h"

#define NSS_DB sp
#define NSS_FILES "shadow"
#define NSS_RES spwd

/* Parse line into result, hold strings at buffer.
   Encode usernames with several variants, selected by scantry
   Return values:
	 2: Line parsed, call more with other variants
	 1: Line parsed, no more variants
	 0: Line correct, but no space in buffer
	-1: Line is incorrect
 */
static int parse_line(struct NSS_RES * result,
			   char * buffer, size_t buflen,
			   char * line, size_t linelen,
			   char * vserver, int vserverlen,
			   int scantry)
{
  int colons = 0, i;
  char *colon[8];
  for (i=0; i<linelen; i++) {
	if (line[i]==':') {
	  if (8==colons) return -1;
	  colon[colons++] = &line[i];
	}
  }
  if (colons<8) return -1;
  {
	int retval = 2;
	int i;
	int bufrest = buflen;
	char * err, * bpos = buffer;
	#define APPEND(a,b) {if(b>=bufrest)return 0;memcpy(bpos,a,b);bpos+=b;bufrest-=b;}
	#define APPENDC(c)  {if(!bufrest--)return 0;*bpos=c; bpos++;}
	#define GETINT(x)   (*(x)==':'?-1:strtol(x, &err, 10)); if (*err != ':') return -1;
	#define GETLINT(x)  (*(x)<'0'&&*(x)>'9'?-1:strtol(x, &err, 10));

	result->sp_namp = bpos;
	  switch(scantry) {
		case 0: // uname@vserver
		  APPEND(line, colon[0]-line);
		  APPENDC('@');
		  APPEND(vserver, vserverlen);
		  break;
		default: // vserver-uname
		  APPEND(vserver, vserverlen);
		  APPENDC('-');
		  APPEND(line, colon[0]-line);
		  retval = 1;
	  }
	  APPENDC(0);
	result->sp_pwdp = bpos;
	  APPEND(colon[0]+1, colon[1]-colon[0]-1);
	  APPENDC(0);

	result->sp_lstchg = GETINT(colon[1]+1)
	result->sp_min    = GETINT(colon[2]+1)
	result->sp_max    = GETINT(colon[3]+1)
	result->sp_warn   = GETINT(colon[4]+1)
	result->sp_inact  = GETINT(colon[5]+1)
	result->sp_expire = GETINT(colon[6]+1)
	result->sp_flag   = GETLINT(colon[7]+1)
	
	return retval;

	#undef APPEND
	#undef APPENDC
	#undef GETINT
	#undef GETLINT
  }
}

/* Select vserver from username */
/* -1 if not vserver linkname */
static int parse_user(char * name,
			   char ** vserver, int * vserverlen,
			   char ** uname, int * unamelen
			  )
{
  int i;
  for (i=0; name[i]; i++) {
	if (name[i]=='@') {
	  int j = strlen(name);
	  *vserver	  = &name[i+1];
	  *vserverlen = j-i-1;
	  *uname	  = name;
	  *unamelen   = i;
	  return 0;
	}
	if (name[i]=='-') {
	  int j = strlen(name);
	  *vserver	  = name;
	  *vserverlen = i;
	  *uname	  = &name[i+1];
	  *unamelen   = j-i-1;
	  return 0;
	}
  }
  return -1;
}

/* Parse line into result, with search of username, that already at result.
   Return values:
	 1: Line parsed fine
	 0: Line fine, but no space available
	-1: Line is incorrect / not one searched
 */
static int parse_line_cmp(struct NSS_RES * result,
				   char * buffer, size_t buflen,
				   char * line, size_t linelen,
				   char * vserver, int vserverlen,
				   char * uname, int unamelen)
{
  int colons = 0, i;
  char *err, *colon[8];
  for (i=0; i<linelen; i++) {
	if (line[i]==':') {
	  if (8==colons) return -1;
	  colon[colons++] = &line[i];
	}
  }
  if (colons<8) return -1;
  if (colon[0]-line != unamelen) return -1;
  if (memcmp(line, uname, unamelen)) return -1;

  {
	char * bpos = buffer + strlen(buffer) + 1;
	int bufrest = buflen - (bpos-buffer);
	#define APPEND(a,b) {if(b>=bufrest)return 0;memcpy(bpos,a,b);bpos+=b;bufrest-=b;}
	#define APPENDC(c)	{if(!bufrest--)return 0;*bpos=c; bpos++;}
	#define GETINT(x)   (*(x)==':'?-1:strtol(x, &err, 10)); if (*err != ':') return -1;
	#define GETLINT(x)  (*(x)<'0'&&*(x)>'9'?-1:strtol(x, &err, 10));

	result->sp_namp = buffer;
	result->sp_pwdp = bpos;
	  APPEND(colon[0]+1, colon[1]-colon[0]-1);
	  APPENDC(0);

	result->sp_lstchg = GETINT(colon[1]+1)
	result->sp_min    = GETINT(colon[2]+1)
	result->sp_max    = GETINT(colon[3]+1)
	result->sp_warn   = GETINT(colon[4]+1)
	result->sp_inact  = GETINT(colon[5]+1)
	result->sp_expire = GETINT(colon[6]+1)
	result->sp_flag   = GETLINT(colon[7]+1)
	
	return 1;

	#undef APPEND
	#undef APPENDC
	#undef GETINT
	#undef GETLINT
  }
}

#include "shared.h"
